/* GmbSprite - a widget for Gnome Memory Blocks
 * (http://gnome-memblocks.sourceforge.net)
 *
 * Copyright (C) 2000 The Free Software Foundation
 *
 * Author: Volodymyr Babin <vb@dwuj.ichf.edu.pl>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>
#include "gmb-sprite-animator.h"

/* Frame holding structure
 */

typedef struct _GmbAnimationFrame GmbAnimationFrame;

struct _GmbAnimationFrame
{
  GmbAnimationFrame * prev;
  GmbAnimationFrame * next;
  GdkPixmap * pixmap;
  GdkBitmap * mask;
  guint16 x_offset;
  guint16 y_offset;
  guint delay_time;
};

static void
gmb_animation_frame_ref_internals (GmbAnimationFrame * frame)
{
  g_return_if_fail (frame != NULL);

  if (frame->pixmap)
    gdk_pixmap_ref (frame->pixmap);
  if (frame->mask)
    gdk_bitmap_ref (frame->mask);
}

static void
gmb_animation_frame_unref_internals (GmbAnimationFrame * frame)
{
  g_return_if_fail (frame != NULL);

  if (frame->pixmap)
    {
      gdk_pixmap_unref (frame->pixmap);
      frame->pixmap = NULL;
    }
  if (frame->mask)
    {
      gdk_bitmap_unref (frame->mask);
      frame->mask = NULL;
    }
}

static inline GmbAnimationFrame *
gmb_animation_frame_create (GdkPixmap * pmap,
			    GdkBitmap * mask,
			    guint16 x_offset,
			    guint16 y_offset,
			    guint delay)
{
  GmbAnimationFrame * frame;

  frame = g_new (GmbAnimationFrame, 1);
  frame->pixmap = pmap;
  frame->mask = mask;
  frame->x_offset = x_offset;
  frame->y_offset = y_offset;
  frame->delay_time = delay;

  return frame;
}

/* Animation container
 */

struct _GmbAnimationPrivat
{
  guint bbox_width, bbox_height;
  GmbAnimationFrame * first_frame;
  GmbAnimationFrame * last_frame;
};

static GtkObjectClass * gmb_animation_parent_class = NULL;

enum
{
  GMB_ANIMATION_BOUND_BOX_CHANGED,
  GMB_ANIMATION_LAST_SIGNAL
};

static guint gmb_animation_signals[GMB_ANIMATION_LAST_SIGNAL] = {0};

static void
gmb_animation_destroy (GtkObject * obj)
{
  GmbAnimation * animation;

  g_return_if_fail (IS_GMB_ANIMATION (obj));

  animation = (GmbAnimation *) obj;
  if (animation->priv)
    {
      GmbAnimationFrame * frame;
      GmbAnimationFrame * next_frame;

      frame = animation->priv->first_frame;
      while (frame)
        {
	  next_frame = frame->next;
	  gmb_animation_frame_unref_internals (frame);
	  g_free (frame);
	  frame = next_frame;
	} 
      g_free (animation->priv);
      animation->priv = NULL;
    }

  if (GTK_OBJECT_CLASS (gmb_animation_parent_class)->destroy)
    (* GTK_OBJECT_CLASS (gmb_animation_parent_class)->destroy) (obj);
}

static void
gmb_animation_init (GmbAnimation * animation)
{
  g_return_if_fail (IS_GMB_ANIMATION (animation));

  animation->priv = g_new0 (GmbAnimationPrivat, 1);
}

static void
gmb_animation_class_init (GmbAnimationClass * klass)
{
  GtkObjectClass * object_class;

  g_return_if_fail (IS_GMB_ANIMATION_CLASS (klass));

  gmb_animation_parent_class = gtk_type_class (GTK_TYPE_OBJECT);

  object_class = (GtkObjectClass *) klass;
  object_class->destroy = gmb_animation_destroy;

  klass->bound_box_changed = NULL;

  gmb_animation_signals[GMB_ANIMATION_BOUND_BOX_CHANGED] =
		gtk_signal_new ("bound_box_changed",
				GTK_RUN_LAST,
				object_class->type,
				GTK_SIGNAL_OFFSET (GmbAnimationClass, bound_box_changed),
				gtk_marshal_NONE__INT_INT,
				GTK_TYPE_NONE,
				2,
				GTK_TYPE_UINT,
				GTK_TYPE_UINT);
  gtk_object_class_add_signals (object_class,
				gmb_animation_signals,
				GMB_ANIMATION_LAST_SIGNAL);
}

GtkType
gmb_animation_get_type (void)
{
  static GtkType gmb_animation_type = 0;
  if (!gmb_animation_type)
    {
      static const GtkTypeInfo gmb_animation_type_info =
      {
	"GmbAnimation",
	sizeof (GmbAnimation),
	sizeof (GmbAnimationClass),
	(GtkClassInitFunc) gmb_animation_class_init,
	(GtkObjectInitFunc) gmb_animation_init,
	NULL, NULL,
	(GtkClassInitFunc) NULL
      };
      gmb_animation_type = gtk_type_unique (GTK_TYPE_OBJECT, &gmb_animation_type_info);
    }
  return gmb_animation_type;
}

GmbAnimation *
gmb_animation_new (void)
{
  return GMB_ANIMATION (gtk_type_new (GMB_ANIMATION_TYPE));
}

void
gmb_animation_append_frame (GmbAnimation * animation,
			    GdkPixmap * pmap,
			    GdkBitmap * mask,
			    guint16 x_offset,
			    guint16 y_offset,
			    guint delay_time)
{
  GmbAnimationFrame * frame;
  guint new_width, new_height;

  g_return_if_fail (IS_GMB_ANIMATION (animation));
  g_return_if_fail (pmap != NULL);

  frame = gmb_animation_frame_create (pmap, mask,
				      x_offset, y_offset,
				      delay_time);
  gmb_animation_frame_ref_internals (frame);

  frame->next = NULL;
  if (animation->priv->last_frame == NULL)
    {
      animation->priv->first_frame = frame;
      animation->priv->last_frame = frame;
      frame->prev = NULL;
    }
  else
    {
      frame->prev = animation->priv->last_frame;
      animation->priv->last_frame->next = frame;
      animation->priv->last_frame = frame;
    }

  gdk_window_get_size (frame->pixmap, &new_width, &new_height);
  new_width += x_offset;
  new_height += y_offset;

  if (new_width > animation->priv->bbox_width ||
      new_height > animation->priv->bbox_height)
    {
      animation->priv->bbox_width = new_width;
      animation->priv->bbox_height = new_height;
      gtk_object_ref (GTK_OBJECT (animation));
      gtk_signal_emit (GTK_OBJECT (animation),
    		       gmb_animation_signals[GMB_ANIMATION_BOUND_BOX_CHANGED],
		       new_width, new_height);
      gtk_object_unref (GTK_OBJECT (animation));
    }
}

void
gmb_animation_prepend_frame (GmbAnimation * animation,
			     GdkPixmap * pmap,
			     GdkBitmap * mask,
			     guint16 x_offset,
			     guint16 y_offset,
			     guint delay_time)
{
  GmbAnimationFrame * frame;
  guint new_width, new_height;

  g_return_if_fail (IS_GMB_ANIMATION (animation));
  g_return_if_fail (pmap != NULL);

  frame = gmb_animation_frame_create (pmap, mask,
				      x_offset, y_offset,
				      delay_time);
  gmb_animation_frame_ref_internals (frame);

  frame->prev = NULL;
  if (animation->priv->first_frame == NULL)
    {
      animation->priv->first_frame = frame;
      animation->priv->last_frame = frame;
      frame->next = NULL;
    }
  else
    {
      frame->next = animation->priv->first_frame;
      animation->priv->first_frame->prev = frame;
      animation->priv->first_frame = frame;
    }

  gdk_window_get_size (frame->pixmap, &new_width, &new_height);
  new_width += x_offset;
  new_height += y_offset;

  if (new_width > animation->priv->bbox_width ||
      new_height > animation->priv->bbox_height)
    {
      animation->priv->bbox_width = new_width;
      animation->priv->bbox_height = new_height;
      gtk_object_ref (GTK_OBJECT (animation));
      gtk_signal_emit (GTK_OBJECT (animation),
    		       gmb_animation_signals[GMB_ANIMATION_BOUND_BOX_CHANGED],
		       new_width, new_height);
      gtk_object_unref (GTK_OBJECT (animation));
    }
}

guint
gmb_animation_get_n_frames (GmbAnimation * animation)
{
  guint result = 0;
  GmbAnimationFrame * tmp;

  g_return_val_if_fail (IS_GMB_ANIMATION (animation), 0);

  for (tmp = animation->priv->first_frame; tmp; tmp = tmp->next)
    result += 1;

  return result;
}

/*  Converter function GdkPixbufAnimation -> GmbAnimation
 */

GmbAnimation *
gmb_animation_new_from_gdk_pixbuf_animation (GdkPixbufAnimation * pb_animation)
{
  GmbAnimation * animation;
  GdkPixbufFrame * pb_frame;
  GList * pb_frames;
  GdkPixbuf * buffer;
  guint depth;
  gint x_offset_prev = 0;
  gint y_offset_prev = 0;
  guint frame_width_prev = 0;
  guint frame_height_prev = 0;

  g_return_val_if_fail (pb_animation != NULL, NULL);

  animation = GMB_ANIMATION (gtk_type_new (gmb_animation_get_type ()));
  depth = gdk_rgb_get_visual ()->depth;

  buffer = gdk_pixbuf_new (GDK_COLORSPACE_RGB,
			   TRUE,
			   8,
			   gdk_pixbuf_animation_get_width (pb_animation),
			   gdk_pixbuf_animation_get_height (pb_animation));

  pb_frames = gdk_pixbuf_animation_get_frames (pb_animation);

  while (pb_frames)
    {
      GdkGC * gc;
      GdkPixmap * pixmap = NULL;
      GdkBitmap * mask = NULL;
      GdkPixbuf * frame_pixbuf;
      gint x_offset, y_offset;
      guint frame_width, frame_height;

      pb_frame = (GdkPixbufFrame *) pb_frames->data;

      x_offset = gdk_pixbuf_frame_get_x_offset (pb_frame);
      y_offset = gdk_pixbuf_frame_get_y_offset (pb_frame);
      frame_pixbuf = gdk_pixbuf_frame_get_pixbuf (pb_frame);
      frame_width = gdk_pixbuf_get_width (frame_pixbuf);
      frame_height = gdk_pixbuf_get_height (frame_pixbuf);

      switch (gdk_pixbuf_frame_get_action (pb_frame))
        {
	  case GDK_PIXBUF_FRAME_RETAIN:
	    if (pb_frames->prev)
	      {
	        gdk_pixbuf_composite (frame_pixbuf, buffer,
				      x_offset, y_offset,
				      frame_width, frame_height,
				      (double) x_offset, (double) y_offset, 1.0, 1.0,
				      GDK_INTERP_HYPER, 255);
		pixmap = gdk_pixmap_new (NULL, frame_width_prev, frame_height_prev, depth);
    		gc = gdk_gc_new (pixmap);
    		gdk_pixbuf_render_to_drawable (buffer, pixmap, gc,
    				               x_offset_prev, y_offset_prev,
				               0, 0,
				               frame_width_prev, frame_height_prev,
				               GDK_RGB_DITHER_NONE, 0, 0);
    		gdk_gc_unref (gc);
    		mask = gdk_pixmap_new (NULL, frame_width_prev, frame_height_prev, 1);
    		gdk_pixbuf_render_threshold_alpha (buffer, mask,
    						   x_offset_prev, y_offset_prev,
						   0, 0,
						   frame_width_prev, frame_height_prev,
						   127);
    		gmb_animation_append_frame (animation,
    					    pixmap,
					    mask,
					    (guint16) x_offset_prev,
					    (guint16) y_offset_prev,
					    (guint) gdk_pixbuf_frame_get_delay_time (pb_frame));
    		gdk_pixmap_unref (pixmap);
    		gdk_bitmap_unref (mask);
		break;
	      }
	  case GDK_PIXBUF_FRAME_DISPOSE:
	    gdk_pixbuf_copy_area (frame_pixbuf,
				  0, 0,
				  frame_width, frame_height,
				  buffer,
				  x_offset, y_offset);
	    pixmap = gdk_pixmap_new (NULL, frame_width, frame_height, depth);
    	    gc = gdk_gc_new (pixmap);
    	    gdk_pixbuf_render_to_drawable (buffer, pixmap, gc,
    				           x_offset, y_offset,
				           0, 0,
				           frame_width, frame_height,
				           GDK_RGB_DITHER_NONE, 0, 0);
            gdk_gc_unref (gc);
    	    mask = gdk_pixmap_new (NULL, frame_width, frame_height, 1);
    	    gdk_pixbuf_render_threshold_alpha (buffer, mask,
    					       x_offset, y_offset,
					       0, 0,
					       frame_width, frame_height,
					       127);
    	    gmb_animation_append_frame (animation,
    				        pixmap,
				        mask,
				        (guint16) x_offset,
				        (guint16) y_offset,
				        (guint) gdk_pixbuf_frame_get_delay_time (pb_frame));
    	    gdk_pixmap_unref (pixmap);
    	    gdk_bitmap_unref (mask);
            x_offset_prev = x_offset;
            y_offset_prev = y_offset;
            frame_width_prev = frame_width;
            frame_height_prev = frame_height;
	    break;
	  case GDK_PIXBUF_FRAME_REVERT:
	    break;
	  default:
	    g_assert_not_reached ();
	}
      pb_frames = pb_frames->next;
    }
  gdk_pixbuf_unref (buffer);
  return animation;
}

/*  Animator (Sprite to be used with GmbSpriteArea)
 */

struct _GmbSpriteAnimatorPrivat
{
  GmbAnimation * animation;
  GmbAnimationFrame * current_frame;
  guint timeout_id;
  guint animation_signal_id;
  GmbSpriteAnimatorLoopType loop_type;
  gdouble speed;
  gint reversed:1;
};

static GmbSpriteClass * gmb_sprite_animator_parent_class = NULL;

enum
{
  GMB_SPRITE_ANIMATOR_PLAYBACK_FINISHED,
  GMB_SPRITE_ANIMATOR_LAST_SIGNAL
};

static guint gmb_sprite_animator_signals[GMB_SPRITE_ANIMATOR_LAST_SIGNAL] = {0};

static void
gmb_sprite_animator_destroy (GtkObject * obj)
{
  GmbSpriteAnimator * animator;

  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (obj));

  animator = (GmbSpriteAnimator *) obj;
  if (animator->priv)
    {
      if (animator->priv->animation)
	{
	  gmb_sprite_animator_stop_playback (animator);
    	  gtk_signal_disconnect (GTK_OBJECT (animator->priv->animation), animator->priv->animation_signal_id);
	  gtk_object_unref (GTK_OBJECT (animator->priv->animation));
	}
      g_free (animator->priv);
      animator->priv = NULL;
    }

  if (GTK_OBJECT_CLASS (gmb_sprite_animator_parent_class)->destroy)
    (* GTK_OBJECT_CLASS (gmb_sprite_animator_parent_class)->destroy) (obj);
}

static void
gmb_sprite_animator_size_request (GmbSprite * sprite, guint16 * width, guint16 * height)
{
  GmbSpriteAnimator * animator;

  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (sprite));

  animator = (GmbSpriteAnimator *) sprite;
  if (animator->priv->animation)
    {
      *width = animator->priv->animation->priv->bbox_width;
      *height = animator->priv->animation->priv->bbox_height;
    }
  else
    {
      *width = 0;
      *height = 0;
    }
}

static void
gmb_sprite_animator_paint (GmbSprite * sprite, GdkDrawable * buffer,
			   GdkRectangle * area, GdkRectangle * buffer_area)
{
  GmbSpriteAnimator * animator;

  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (sprite));

  animator = (GmbSpriteAnimator *) sprite;
  if (animator->priv->current_frame)
    {
      GdkGC * gc;

      gc = gdk_gc_new (buffer);
      if (animator->priv->current_frame->mask)
	{
          gdk_gc_set_clip_origin (gc,
		sprite->area.x - buffer_area->x + animator->priv->current_frame->x_offset,
	        sprite->area.y - buffer_area->y + animator->priv->current_frame->y_offset);
          gdk_gc_set_clip_mask (gc, animator->priv->current_frame->mask);
	}
      gdk_draw_pixmap (buffer,
    		       gc,
		       animator->priv->current_frame->pixmap,
		       area->x - sprite->area.x,
		       area->y - sprite->area.y,
		       area->x - buffer_area->x + animator->priv->current_frame->x_offset,
	    	       area->y - buffer_area->y + animator->priv->current_frame->x_offset,
		       area->width, area->height);
      gdk_gc_unref (gc);
    }
}

static void
gmb_sprite_animator_request_update (GmbSpriteAnimator * animator,
				    GmbAnimationFrame * new_frame)
{
  GmbSprite * sprite;

  sprite = (GmbSprite *) animator;

  if (sprite->host == NULL)
    return;

  if (new_frame == NULL)
    gmb_sprite_area_request_update (sprite->host, &sprite->area);
  else
    {
      gint16 x_real_offset;
      gint16 y_real_offset;
      GdkRectangle update;

      x_real_offset = MIN (new_frame->x_offset, animator->priv->current_frame->x_offset);
      y_real_offset = MIN (new_frame->y_offset, animator->priv->current_frame->y_offset);

      update.x = sprite->area.x + x_real_offset;
      update.y = sprite->area.y + y_real_offset;
      update.width = sprite->area.width - x_real_offset;
      update.height = sprite->area.height - y_real_offset;
      gmb_sprite_area_request_update (sprite->host, &update);
    }
}

static gint
gmb_sprite_animator_animator_cb (GmbSpriteAnimator * animator)
{
  GmbAnimationFrame * next_frame = NULL;

  if (animator->priv->reversed)
    next_frame = animator->priv->current_frame->prev;
  else
    next_frame = animator->priv->current_frame->next;

  if (next_frame == NULL)
    {
      switch (animator->priv->loop_type)
        {
          case GMB_SPRITE_ANIMATOR_LOOP_NONE :
            animator->priv->timeout_id = 0;
	    gtk_object_ref (GTK_OBJECT (animator));
            gtk_signal_emit (GTK_OBJECT (animator), gmb_sprite_animator_signals[GMB_SPRITE_ANIMATOR_PLAYBACK_FINISHED]);
	    gtk_object_unref (GTK_OBJECT (animator));
	    return 0;
	    break;
          case GMB_SPRITE_ANIMATOR_LOOP_RESTART :
	    next_frame = (animator->priv->reversed ? animator->priv->animation->priv->last_frame :
						     animator->priv->animation->priv->first_frame);
	    break;
          case GMB_SPRITE_ANIMATOR_LOOP_PING_PONG :
	    animator->priv->reversed ^= 1;
	    next_frame = animator->priv->current_frame;
	    break;
	  default :
	    g_assert_not_reached ();
	}
    }

  gmb_sprite_animator_request_update (animator, next_frame);
  animator->priv->current_frame = next_frame;
  animator->priv->timeout_id =
  gtk_timeout_add ((guint) floor (((gdouble) animator->priv->current_frame->delay_time)*(animator->priv->speed + 0.5)),
		   (GtkFunction) gmb_sprite_animator_animator_cb, animator);
  return 0;
}

static void
gmb_sprite_animator_animation_destroy_cb (GmbSpriteAnimator * animator, gpointer data)
{
  g_assert (IS_GMB_SPRITE_ANIMATOR (animator));

  gmb_sprite_animator_stop_playback (animator);
  animator->priv->animation = NULL;
  animator->priv->current_frame = NULL;
  gmb_sprite_request_resize ((GmbSprite *) animator);
}

static void
gmb_sprite_animator_init (GmbSpriteAnimator * animator)
{
  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (animator));

  animator->priv = g_new0 (GmbSpriteAnimatorPrivat, 1);
  animator->priv->loop_type = GMB_SPRITE_ANIMATOR_LOOP_RESTART;
  animator->priv->speed = 1.0;
}

static void
gmb_sprite_animator_class_init (GmbSpriteAnimatorClass * klass)
{
  GtkObjectClass * object_class;
  GmbSpriteClass * sprite_class;

  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR_CLASS (klass));

  gmb_sprite_animator_parent_class = gtk_type_class (GMB_SPRITE_TYPE);

  object_class = (GtkObjectClass *) klass;
  object_class->destroy = gmb_sprite_animator_destroy;

  sprite_class = (GmbSpriteClass *) klass;
  sprite_class->size_request = gmb_sprite_animator_size_request;
  sprite_class->paint = gmb_sprite_animator_paint;

  klass->playback_finished = NULL;

  gmb_sprite_animator_signals[GMB_SPRITE_ANIMATOR_PLAYBACK_FINISHED] =
		gtk_signal_new ("playback_finished",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (GmbSpriteAnimatorClass, playback_finished),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class,
				gmb_sprite_animator_signals,
				GMB_SPRITE_ANIMATOR_LAST_SIGNAL);
}

GtkType
gmb_sprite_animator_get_type (void)
{
  static GtkType gmb_animator_type = 0;
  if (!gmb_animator_type)
    {
      static const GtkTypeInfo gmb_animator_type_info =
      {
	"GmbSpriteAnimator",
	sizeof (GmbSpriteAnimator),
	sizeof (GmbSpriteAnimatorClass),
	(GtkClassInitFunc) gmb_sprite_animator_class_init,
	(GtkObjectInitFunc) gmb_sprite_animator_init,
	NULL, NULL,
	(GtkClassInitFunc) NULL
      };
      gmb_animator_type = gtk_type_unique (GMB_SPRITE_TYPE, &gmb_animator_type_info);
    }
  return gmb_animator_type;
}

GmbSprite *
gmb_sprite_animator_new (void)
{
  return GMB_SPRITE (gtk_type_new (GMB_SPRITE_ANIMATOR_TYPE));
}

void
gmb_sprite_animator_set_animation (GmbSpriteAnimator * animator, GmbAnimation * new_animation)
{
  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (animator));

  if (animator->priv->animation)
    {
      gmb_sprite_animator_stop_playback (animator);
      gtk_signal_disconnect (GTK_OBJECT (animator->priv->animation), animator->priv->animation_signal_id);
      gtk_object_unref (GTK_OBJECT (animator->priv->animation));
      animator->priv->animation = NULL;
      animator->priv->current_frame = NULL;
    }

  animator->priv->animation = new_animation;

  if (new_animation)
    {
      gtk_object_ref (GTK_OBJECT (new_animation));
      gtk_object_sink (GTK_OBJECT (new_animation));

      animator->priv->animation_signal_id =
        gtk_signal_connect_object (GTK_OBJECT (new_animation),
    				   "bound_box_changed",
			           GTK_SIGNAL_FUNC (gmb_sprite_request_resize),
			           GTK_OBJECT (animator));
      gtk_signal_connect_object_while_alive (GTK_OBJECT (new_animation),
					     "destroy",
					     GTK_SIGNAL_FUNC (gmb_sprite_animator_animation_destroy_cb),
					     GTK_OBJECT (animator));
      animator->priv->current_frame = new_animation->priv->first_frame;
    }

  gmb_sprite_request_resize ((GmbSprite *) animator);
}

const GmbAnimation *
gmb_sprite_animator_get_animation (GmbSpriteAnimator * animator)
{
  g_return_val_if_fail (IS_GMB_SPRITE_ANIMATOR (animator), NULL);

  return animator->priv->animation;
}

gboolean
gmb_sprite_animator_is_reversed (const GmbSpriteAnimator * animator)
{
  g_return_val_if_fail (IS_GMB_SPRITE_ANIMATOR (animator), FALSE);

  return (gboolean) animator->priv->reversed;
}

void gmb_sprite_animator_set_reversed (GmbSpriteAnimator * animator, gboolean reversed)
{
  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (animator));

  animator->priv->reversed = reversed;
}

gboolean
gmb_sprite_animator_is_running (const GmbSpriteAnimator * animator)
{
  g_return_val_if_fail (IS_GMB_SPRITE_ANIMATOR (animator), FALSE);

  return (animator->priv->timeout_id == 0 ? FALSE : TRUE);
}

void
gmb_sprite_animator_start_playback (GmbSpriteAnimator * animator)
{
  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (animator));

  if (animator->priv->timeout_id)
    return;

  if (animator->priv->current_frame == NULL)
    {
      gtk_object_ref (GTK_OBJECT (animator));
      gtk_signal_emit (GTK_OBJECT (animator), gmb_sprite_animator_signals[GMB_SPRITE_ANIMATOR_PLAYBACK_FINISHED]);
      gtk_object_unref (GTK_OBJECT (animator));
    }
  else
    gmb_sprite_animator_animator_cb (animator);
}

void
gmb_sprite_animator_stop_playback (GmbSpriteAnimator * animator)
{
  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (animator));

  if (animator->priv->timeout_id)
    {
      gtk_timeout_remove (animator->priv->timeout_id);
      animator->priv->timeout_id = 0;
    }
}

GmbSpriteAnimatorLoopType
gmb_sprite_animator_get_loop_type (const GmbSpriteAnimator * animator)
{
  g_return_val_if_fail (IS_GMB_SPRITE_ANIMATOR (animator), 0);

  return animator->priv->loop_type;
}

void
gmb_sprite_animator_set_loop_type (GmbSpriteAnimator * animator,
				   GmbSpriteAnimatorLoopType loop_type)
{
  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (animator));
  g_return_if_fail (loop_type >= GMB_SPRITE_ANIMATOR_LOOP_NONE &&
		    loop_type <= GMB_SPRITE_ANIMATOR_LOOP_PING_PONG);

  animator->priv->loop_type = loop_type;
}

gdouble
gmb_sprite_animator_get_playback_speed (const GmbSpriteAnimator * animator)
{
  g_return_val_if_fail (IS_GMB_SPRITE_ANIMATOR (animator), 0.0);

  return animator->priv->speed;
}

void
gmb_sprite_animator_set_playback_speed (GmbSpriteAnimator * animator, gdouble speed)
{
  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (animator));
  g_return_if_fail (speed >= 0.0);

  animator->priv->speed = speed;
}

guint
gmb_sprite_animator_get_animation_n_frames (GmbSpriteAnimator * animator)
{
  g_return_val_if_fail (IS_GMB_SPRITE_ANIMATOR (animator), 0);

  if (animator->priv->animation)
    return gmb_animation_get_n_frames (animator->priv->animation);
  else
    return 0;
}

void
gmb_sprite_animator_advance (GmbSpriteAnimator * animator, gint how_many)
{
  gint tmp;
  GmbAnimationFrame * new_current;

  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (animator));

  if (animator->priv->current_frame == NULL || how_many == 0)
    return;

  new_current = animator->priv->current_frame;
  for (tmp = 0; tmp < ABS (how_many); tmp++)
    {
      GmbAnimationFrame * candidate;

      candidate = (how_many > 0 ? new_current->next : new_current->prev);
      if (candidate)
        new_current = candidate;
      else
        break;
    }

  gmb_sprite_animator_request_update (animator, new_current);
  animator->priv->current_frame = new_current;
}

void
gmb_sprite_animator_goto_begin (GmbSpriteAnimator * animator)
{
  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (animator));

  if (animator->priv->animation)
    animator->priv->current_frame = animator->priv->animation->priv->first_frame;
}

void
gmb_sprite_animator_goto_end (GmbSpriteAnimator * animator)
{
  g_return_if_fail (IS_GMB_SPRITE_ANIMATOR (animator));

  if (animator->priv->animation)
    animator->priv->current_frame = animator->priv->animation->priv->last_frame;
}
