/* GmbSprite - a widget for Gnome Memory Blocks
 * (http://gnome-memblocks.sourceforge.net)
 *
 * Copyright (C) 2000 The Free Software Foundation
 *
 * Author: Volodymyr Babin <vb@dwuj.ichf.edu.pl>
 *         Lots of GnomeCanvas events handling code has been used.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gtk/gtkmain.h>
#include <gtk/gtkstyle.h>
#include <gtk/gtksignal.h>
#include "gmb-sprite-area.h"
#include "gmb-sprite-utils.h"

/*
 *  GMB Sprite Abstract Class
 */

struct _GmbSpritePrivat
{
  GmbSprite * prev;
  GmbSprite * next;
};

enum
{
  GMB_SPRITE_EVENT,
  GMB_SPRITE_LAST_SIGNAL
};

static guint gmb_sprite_signals[GMB_SPRITE_LAST_SIGNAL] = {0};
static GtkObjectClass * gmb_sprite_parent_class = NULL;

static void
gmb_sprite_shutdown (GtkObject * object)
{
  GmbSprite * sprite;

  g_return_if_fail (IS_GMB_SPRITE (object));

  sprite = (GmbSprite *) object;

  if (sprite->priv)
    {
      gmb_sprite_unparent (sprite);
      g_free (sprite->priv);
      sprite->priv = NULL;
    }

  if (GTK_OBJECT_CLASS (gmb_sprite_parent_class)->shutdown)
    (* GTK_OBJECT_CLASS (gmb_sprite_parent_class)->shutdown) (object);
}

typedef gint (* GmbSpriteSignal1) (GtkObject * sprite, gpointer arg1, gpointer data);

static void
gmb_sprite_marshal_signal_1 (GtkObject *object,
			     GtkSignalFunc func,
			     gpointer func_data,
			     GtkArg *args)
{
  GmbSpriteSignal1 rfunc;
  gint *return_val;

  rfunc = (GmbSpriteSignal1) func;
  return_val = GTK_RETLOC_BOOL (args[1]);

  *return_val = (* rfunc) (object,
			   GTK_VALUE_BOXED (args[0]),
			   func_data);
}

static void
gmb_sprite_init (GmbSprite * sprite)
{
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  sprite->priv = g_new0 (GmbSpritePrivat, 1);
  sprite->object.flags |= GMB_SPRITE_VISIBLE;
}

static void
gmb_sprite_class_init (GmbSpriteClass * sprite_class)
{
  GtkObjectClass * object_class;

  g_return_if_fail (IS_GMB_SPRITE_CLASS (sprite_class));

  gmb_sprite_parent_class = gtk_type_class (GTK_TYPE_OBJECT);

  object_class = (GtkObjectClass *) sprite_class;
  object_class->shutdown = gmb_sprite_shutdown;

  gmb_sprite_signals[GMB_SPRITE_EVENT] =
	gtk_signal_new ("event",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET (GmbSpriteClass, event),
			gmb_sprite_marshal_signal_1,
			GTK_TYPE_BOOL, 1,
			GTK_TYPE_GDK_EVENT);

  gtk_object_class_add_signals (object_class,
				gmb_sprite_signals,
				GMB_SPRITE_LAST_SIGNAL);
}

GtkType
gmb_sprite_get_type (void)
{
  static GtkType sprite_type = 0;

  if (!sprite_type)
    {
      static const GtkTypeInfo sprite_type_info =
      {
	"GmbSprite",
	sizeof (GmbSprite),
	sizeof (GmbSpriteClass),
	(GtkClassInitFunc) gmb_sprite_class_init,
	(GtkObjectInitFunc) gmb_sprite_init,
	NULL, NULL,
	(GtkClassInitFunc) NULL
      };
      sprite_type = gtk_type_unique (GTK_TYPE_OBJECT, &sprite_type_info);
    }
  return sprite_type;
}

/*
 *  GMB Sprite Area
 */

struct _GmbSpriteAreaPrivat
{
  GSList * update_areas;
  GdkGC * gc;
  /* list of sprites */
  GmbSprite * bottom;
  GmbSprite * top;
  /* event handling */
  GmbSprite * current;
  GmbSprite * new_current;
  GdkEvent pick_event;
  gint state;
  gint event_x, event_y;
  guint need_repick : 1;
  /* redraw related stuff */
  guint idle_id;
  guint frozen : 1;
};

static GtkWidgetClass * gmb_sprite_area_parent_class = NULL;

static void
gmb_sprite_area_realize_items (GmbSpriteArea * sa)
{
  GmbSprite * sprite;

  sprite = sa->priv->bottom;
  while (sprite)
    {
      if (!(sprite->object.flags & GMB_SPRITE_REALIZED))
        {
	  if (GMB_SPRITE_CLASS (sprite->object.klass)->realize)
	    (* GMB_SPRITE_CLASS (sprite->object.klass)->realize) (sprite);

	  sprite->object.flags |= GMB_SPRITE_REALIZED;
	}
      sprite = sprite->priv->next;
    }
}

static void
gmb_sprite_area_unrealize_items (GmbSpriteArea * sa)
{
  GmbSprite * sprite;

  sprite = sa->priv->bottom;
  while (sprite)
    {
      if (sprite->object.flags & GMB_SPRITE_REALIZED)
        {
	  if (GMB_SPRITE_CLASS (sprite->object.klass)->unrealize)
	    (* GMB_SPRITE_CLASS (sprite->object.klass)->unrealize) (sprite);

	  sprite->object.flags &= ~GMB_SPRITE_REALIZED;
	}
      sprite = sprite->priv->next;
    }
}

static void
gmb_sprite_area_cancel_queue (GmbSpriteArea * sa)
{
  g_return_if_fail (IS_GMB_SPRITE_AREA (sa));

  if (sa->priv->idle_id)
    {
      gtk_idle_remove (sa->priv->idle_id);
      sa->priv->idle_id = 0;
    }

  if (sa->priv->update_areas)
    {
      g_slist_foreach (sa->priv->update_areas, (GFunc) g_free, NULL);
      g_slist_free (sa->priv->update_areas);
      sa->priv->update_areas = NULL;
    }
}

static void
gmb_sprite_area_create_gc (GmbSpriteArea * sa)
{
  GtkWidget * widget;
  GtkStyle * style;
  GtkStateType state;

  if (!GTK_WIDGET_REALIZED (sa))
    return;

  if (sa->priv->gc)
    gdk_gc_unref (sa->priv->gc);

  widget = (GtkWidget *) sa;

  style = gtk_widget_get_style (widget);
  state = GTK_WIDGET_STATE (widget);

  sa->priv->gc = gdk_gc_new (widget->window);
  gdk_gc_copy (sa->priv->gc, style->bg_gc[state]);

  if (style->bg_pixmap[state] != NULL)
    {
      gdk_gc_set_tile (sa->priv->gc, style->bg_pixmap[state]);
      gdk_gc_set_fill (sa->priv->gc, GDK_TILED);
    }
  else
    gdk_gc_set_fill (sa->priv->gc, GDK_SOLID);
}

static void
gmb_sprite_area_shutdown (GtkObject * obj)
{
  GmbSpriteArea * sa;

  g_return_if_fail (IS_GMB_SPRITE_AREA (obj));

  sa = (GmbSpriteArea *) obj;
  if (sa->priv)
    {
      gmb_sprite_area_cancel_queue (sa);
      gmb_sprite_area_foreach (sa, (GmbSpriteFunc) gtk_object_destroy, NULL);

      if (sa->priv->gc)
        gdk_gc_unref (sa->priv->gc);

      g_free (sa->priv);
      sa->priv = NULL;
    }

  if (GTK_OBJECT_CLASS (gmb_sprite_area_parent_class)->shutdown)
    (* GTK_OBJECT_CLASS (gmb_sprite_area_parent_class)->shutdown) (obj);
}

static void
gmb_sprite_area_size_allocate (GtkWidget * widget, GtkAllocation * allocation)
{
  g_return_if_fail (IS_GMB_SPRITE_AREA (widget));
  g_return_if_fail (allocation != NULL);

  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (widget))
    gdk_window_move_resize (widget->window,
    			    allocation->x,
			    allocation->y,
			    allocation->width,
			    allocation->height);

  /* we will receive exposition for whole window, so */
  gmb_sprite_area_cancel_queue ((GmbSpriteArea *) widget);
}

static void
gmb_sprite_area_realize (GtkWidget * widget)
{
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (IS_GMB_SPRITE_AREA (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK
			    | GDK_ENTER_NOTIFY_MASK
			    | GDK_LEAVE_NOTIFY_MASK
			    | GDK_POINTER_MOTION_MASK
			    | GDK_BUTTON_PRESS_MASK
			    | GDK_BUTTON_RELEASE_MASK);
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget),
			    	   &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, widget);
  widget->style = gtk_style_attach (widget->style, widget->window);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);
  gmb_sprite_area_create_gc ((GmbSpriteArea *) widget);
  gmb_sprite_area_realize_items ((GmbSpriteArea *) widget);
}

static void
gmb_sprite_area_unrealize (GtkWidget * widget)
{
  GmbSpriteArea * sa;

  g_return_if_fail (IS_GMB_SPRITE_AREA (widget));

  sa = (GmbSpriteArea *) widget;

  gmb_sprite_area_unrealize_items (sa);
  if (sa->priv->gc)
    {
      gdk_gc_unref (sa->priv->gc);
      sa->priv->gc = NULL;
    }

  if (GTK_WIDGET_CLASS (gmb_sprite_area_parent_class)->unrealize)
    (* GTK_WIDGET_CLASS (gmb_sprite_area_parent_class)->unrealize) (widget);
}

static void
gmb_sprite_area_do_update_area (GmbSpriteArea * sa, GdkRectangle * area)
{
  GmbSprite * sprite;
  GdkPixmap * buffer;
  GtkWidget * widget;

  if (!GTK_WIDGET_REALIZED (sa) || !GTK_WIDGET_DRAWABLE (sa))
    return;

  widget = (GtkWidget *) sa;
  buffer = gdk_pixmap_new (widget->window, area->width, area->height, -1);

  gdk_gc_set_ts_origin (sa->priv->gc, - widget->allocation.x - area->x, - widget->allocation.y - area->y);
  gdk_draw_rectangle (buffer, sa->priv->gc, TRUE, 0, 0, area->width, area->height);

  sprite = sa->priv->bottom;
  while (sprite)
    {
      GdkRectangle intersection;

      if ((sprite->object.flags & GMB_SPRITE_VISIBLE) &&
    	  gdk_rectangle_intersect (&sprite->area, area, &intersection) &&
	  GMB_SPRITE_CLASS (sprite->object.klass)->paint)
	  (* GMB_SPRITE_CLASS (sprite->object.klass)->paint) (sprite, buffer, &intersection, area);
      sprite = sprite->priv->next;
    }
  gdk_window_copy_area (widget->window,
			sa->priv->gc,
			area->x, area->y,
			buffer, 0, 0, area->width, area->height);
#if 0
  gdk_draw_rectangle (widget->window,
		      widget->style->black_gc, FALSE,
		      area->x, area->y,
		      area->width, area->height);
#endif
  gdk_flush ();
  gdk_pixmap_unref (buffer);
}

static gboolean
gmb_sprite_area_expose (GtkWidget * widget, GdkEventExpose * event)
{
  g_return_val_if_fail (IS_GMB_SPRITE_AREA (widget) && event != NULL, FALSE);

  gmb_sprite_area_do_update_area ((GmbSpriteArea *) widget, &event->area);
  return FALSE;
}

static gint gmb_sprite_area_pick_current (GmbSpriteArea *, GdkEvent *);

static gint
gmb_sprite_area_idle_func (GmbSpriteArea * sa)
{
  GSList * tmp;

  sa->priv->idle_id = 0;
  tmp = sa->priv->update_areas;
  while (tmp)
    {
      gmb_sprite_area_do_update_area (sa, (GdkRectangle *) tmp->data);
      g_free (tmp->data);
      tmp = tmp->next;
    }
  g_slist_free (sa->priv->update_areas);
  sa->priv->update_areas = NULL;
  if (sa->priv->need_repick)
    {
      sa->priv->need_repick = FALSE;
      gmb_sprite_area_pick_current (sa, &sa->priv->pick_event);
    }
  return 0;
}

static void
gmb_sprite_area_add_area_to_queue (GmbSpriteArea * sa, GdkRectangle * area)
{
  GSList * tmp;

  tmp = sa->priv->update_areas;
  while (tmp)
    {
      GdkRectangle intersection, aunion;

      if (gdk_rectangle_intersect ((GdkRectangle *) tmp->data, area, &intersection))
        {
	  aunion.x = MIN (area->x, ((GdkRectangle *) tmp->data)->x);
	  aunion.y = MIN (area->y, ((GdkRectangle *) tmp->data)->y);
	  aunion.width = MAX (area->x + area->width, ((GdkRectangle *) tmp->data)->x + ((GdkRectangle *) tmp->data)->width) - aunion.x;
	  aunion.height = MAX (area->y + area->height, ((GdkRectangle *) tmp->data)->y + ((GdkRectangle *) tmp->data)->height) - aunion.y;
	  g_free (tmp->data);
	  sa->priv->update_areas = g_slist_remove_link (sa->priv->update_areas, tmp);
	  g_slist_free_1 (tmp);
	  gmb_sprite_area_add_area_to_queue (sa, &aunion);
	  return;
	}
      tmp = tmp->next;
    }
  sa->priv->update_areas = g_slist_append (sa->priv->update_areas,
			    		   g_memdup (area, sizeof (GdkRectangle)));
}

static void
gmb_sprite_area_style_set (GtkWidget * widget, GtkStyle * prev)
{
  g_return_if_fail (IS_GMB_SPRITE_AREA (widget));

  gmb_sprite_area_cancel_queue ((GmbSpriteArea *) widget);
  gmb_sprite_area_create_gc ((GmbSpriteArea *) widget);
}

static void
gmb_sprite_area_state_changed (GtkWidget * widget, GtkStateType prev)
{
  g_return_if_fail (IS_GMB_SPRITE_AREA (widget));

  gmb_sprite_area_cancel_queue ((GmbSpriteArea *) widget);
  gmb_sprite_area_create_gc ((GmbSpriteArea *) widget);
}

static gint
gmb_sprite_area_emit_event (GmbSpriteArea * sa, GdkEvent * event)
{
  gint finished = FALSE;

  if(sa->priv->current)
    {
      gtk_object_ref (GTK_OBJECT (sa->priv->current));
      gtk_signal_emit (GTK_OBJECT (sa->priv->current),
		       gmb_sprite_signals[GMB_SPRITE_EVENT],
		       event,
		       &finished);
      if (GTK_OBJECT_DESTROYED (sa->priv->current))
	finished = TRUE;
      gtk_object_unref (GTK_OBJECT (sa->priv->current));
    }
  return finished;
}

static gint
gmb_sprite_area_pick_current (GmbSpriteArea * sa, GdkEvent * event)
{
  gint button_down;
  gdouble x, y;
  gint retval;

  retval = FALSE;

  /* If a button is down, we'll perform enter and leave events on the
   * current item, but not enter on any other item.  This is more or less
   * like X pointer grabbing for canvas items.
   */
  button_down = sa->priv->state & (GDK_BUTTON1_MASK
				 | GDK_BUTTON2_MASK
			         | GDK_BUTTON3_MASK
			         | GDK_BUTTON4_MASK
			         | GDK_BUTTON5_MASK);
  /* Save the event in the canvas.  This is used to synthesize enter and
   * leave events in case the current item changes.  It is also used to
   * re-pick the current item if the current one gets deleted.  Also,
   * synthesize an enter event.
   */
  if (event != &sa->priv->pick_event)
    {
      if ((event->type == GDK_MOTION_NOTIFY) || (event->type == GDK_BUTTON_RELEASE))
        {
	  /* these fields have the same offsets in both types of events */

	  sa->priv->pick_event.crossing.type       = GDK_ENTER_NOTIFY;
	  sa->priv->pick_event.crossing.window     = event->motion.window;
	  sa->priv->pick_event.crossing.send_event = event->motion.send_event;
	  sa->priv->pick_event.crossing.subwindow  = NULL;
	  sa->priv->pick_event.crossing.x          = event->motion.x;
	  sa->priv->pick_event.crossing.y          = event->motion.y;
	  sa->priv->pick_event.crossing.mode       = GDK_CROSSING_NORMAL;
	  sa->priv->pick_event.crossing.detail     = GDK_NOTIFY_NONLINEAR;
	  sa->priv->pick_event.crossing.focus      = FALSE;
	  sa->priv->pick_event.crossing.state      = event->motion.state;

	  /* these fields don't have the same offsets in both types of events */

        if (event->type == GDK_MOTION_NOTIFY)
          {
	    sa->priv->pick_event.crossing.x_root = event->motion.x_root;
	    sa->priv->pick_event.crossing.y_root = event->motion.y_root;
	  }
        else
          {
	    sa->priv->pick_event.crossing.x_root = event->button.x_root;
	    sa->priv->pick_event.crossing.y_root = event->button.y_root;
	  }
        }
    }
  else
    sa->priv->pick_event = *event;

  /* LeaveNotify means that there is no current item, so we don't look for one */

  if (sa->priv->pick_event.type != GDK_LEAVE_NOTIFY)
    {
      /* these fields don't have the same offsets in both types of events */

      if (sa->priv->pick_event.type == GDK_ENTER_NOTIFY)
        {
	  x = sa->priv->pick_event.crossing.x;
	  y = sa->priv->pick_event.crossing.y;
	}
      else
        {
	  x = sa->priv->pick_event.motion.x;
	  y = sa->priv->pick_event.motion.y;
	}

      /* canvas pixel coords */

      sa->priv->event_x = (gint) (x + 0.5);
      sa->priv->event_y = (gint) (y + 0.5);

      /* find the closest item */
      sa->priv->new_current = NULL;

      if (sa->priv->top)
        {
	  GmbSprite * tmp;
	  tmp = sa->priv->top;
	  while (tmp)
	    {
              if (POINT_INSIDE_RECTANGLE (sa->priv->event_x, sa->priv->event_y, &tmp->area) &&
	          (tmp->object.flags & GMB_SPRITE_VISIBLE))
		{
		  sa->priv->new_current = tmp;
		  break;
		}
	      tmp = tmp->priv->prev;
	    }
	}
    }
  else
    sa->priv->new_current = NULL;

  if (sa->priv->new_current == sa->priv->current)
    return retval; /* current item did not change */

  /* Synthesize events for old and new current items */

  if ((sa->priv->new_current != sa->priv->current) && (sa->priv->current != NULL))
    {
      GdkEvent new_event;

      new_event = sa->priv->pick_event;
      new_event.type = GDK_LEAVE_NOTIFY;

      new_event.crossing.detail = GDK_NOTIFY_ANCESTOR;
      new_event.crossing.subwindow = NULL;
      retval = gmb_sprite_area_emit_event (sa, &new_event);
    }

  /* new_current_item may have been set to NULL during the call to emit_event() above */

  if ((sa->priv->new_current != sa->priv->current) && button_down)
    return retval;

  /* Handle the rest of cases */

  sa->priv->current = sa->priv->new_current;

  if (sa->priv->current != NULL)
    {
      GdkEvent new_event;

      new_event = sa->priv->pick_event;
      new_event.type = GDK_ENTER_NOTIFY;
      new_event.crossing.detail = GDK_NOTIFY_ANCESTOR;
      new_event.crossing.subwindow = NULL;
      retval = gmb_sprite_area_emit_event (sa, &new_event);
    }
  return retval;
}

static gint
gmb_sprite_area_button (GtkWidget * widget, GdkEventButton * event)
{
  GmbSpriteArea * sa;
  gint mask;
  gint retval;

  g_return_val_if_fail (IS_GMB_SPRITE_AREA (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  retval = FALSE;

  if (event->window != widget->window)
    return retval;

  sa = (GmbSpriteArea *) widget;

  if (sa->priv->frozen)
    return FALSE;

  switch (event->button)
    {
      case 1:
	mask = GDK_BUTTON1_MASK;
	break;
      case 2:
	mask = GDK_BUTTON2_MASK;
	break;
      case 3:
	mask = GDK_BUTTON3_MASK;
	break;
      case 4:
	mask = GDK_BUTTON4_MASK;
	break;
      case 5:
	mask = GDK_BUTTON5_MASK;
	break;
      default:
	mask = 0;
    }

  switch (event->type)
    {
      case GDK_BUTTON_PRESS:
      case GDK_2BUTTON_PRESS:
      case GDK_3BUTTON_PRESS:
	/* Pick the current item as if the button were not pressed, and
	 * then process the event.
	 */
	sa->priv->state = event->state;
	gmb_sprite_area_pick_current (sa, (GdkEvent *) event);
	sa->priv->state ^= mask;
	retval = gmb_sprite_area_emit_event (sa, (GdkEvent *) event);
	break;
      case GDK_BUTTON_RELEASE:
	/* Process the event as if the button were pressed, then repick
	 * after the button has been released
	 */
	sa->priv->state = event->state;
	retval = gmb_sprite_area_emit_event (sa, (GdkEvent *) event);
	event->state ^= mask;
	sa->priv->state = event->state;
	gmb_sprite_area_pick_current (sa, (GdkEvent *) event);
	event->state ^= mask;
	break;
      default:
	g_assert_not_reached ();
  }
  return retval;
}

static gint
gmb_sprite_area_motion (GtkWidget * widget, GdkEventMotion * event)
{
  GmbSpriteArea * sa;

  g_return_val_if_fail (IS_GMB_SPRITE_AREA (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->window != widget->window)
    return FALSE;

  sa = (GmbSpriteArea *) widget;

  if (sa->priv->frozen)
    return FALSE;

  sa->priv->state = event->state;
  gmb_sprite_area_pick_current (sa, (GdkEvent *) event);

  return gmb_sprite_area_emit_event (sa, (GdkEvent *) event);
}

static gint
gmb_sprite_area_crossing (GtkWidget * widget, GdkEventCrossing * event)
{
  GmbSpriteArea * sa;

  g_return_val_if_fail (IS_GMB_SPRITE_AREA (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->window != widget->window)
    return FALSE;

  sa = (GmbSpriteArea *) widget;

  if (sa->priv->frozen)
    return FALSE;

  sa->priv->state = event->state;

  return gmb_sprite_area_pick_current (sa, (GdkEvent *) event);
}

static void
gmb_sprite_area_setup_added_sprite (GmbSprite * sprite)
{
  if (GTK_WIDGET_REALIZED (sprite->host) &&
      !(sprite->object.flags & GMB_SPRITE_REALIZED))
    {
      if (GMB_SPRITE_CLASS (sprite->object.klass)->realize)
        (* GMB_SPRITE_CLASS (sprite->object.klass)->realize) (sprite);
      sprite->object.flags |= GMB_SPRITE_REALIZED;
    }
  gmb_sprite_request_resize (sprite);
}

static void
gmb_sprite_area_init (GmbSpriteArea * sa)
{
  g_return_if_fail (IS_GMB_SPRITE_AREA (sa));

  GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (sa), GTK_NO_WINDOW);
  sa->priv = g_new0 (GmbSpriteAreaPrivat, 1);
  sa->priv->need_repick = TRUE;
}

static void
gmb_sprite_area_class_init (GmbSpriteAreaClass * klass)
{
  GtkObjectClass * object_class;
  GtkWidgetClass * widget_class;

  g_return_if_fail (IS_GMB_SPRITE_AREA_CLASS (klass));

  gmb_sprite_area_parent_class = gtk_type_class (GTK_TYPE_WIDGET);

  object_class = (GtkObjectClass *) klass;
  object_class->shutdown = gmb_sprite_area_shutdown;

  widget_class = (GtkWidgetClass *) klass;
  widget_class->size_allocate = gmb_sprite_area_size_allocate;
  widget_class->realize = gmb_sprite_area_realize;
  widget_class->unrealize = gmb_sprite_area_unrealize;
  widget_class->expose_event = gmb_sprite_area_expose;
  widget_class->style_set = gmb_sprite_area_style_set;
  widget_class->state_changed = gmb_sprite_area_state_changed;
  widget_class->button_press_event = gmb_sprite_area_button;
  widget_class->button_release_event = gmb_sprite_area_button;
  widget_class->motion_notify_event = gmb_sprite_area_motion;
  widget_class->enter_notify_event = gmb_sprite_area_crossing;
  widget_class->leave_notify_event = gmb_sprite_area_crossing;
}

GtkType
gmb_sprite_area_get_type (void)
{
  static GtkType sprite_area_type = 0;

  if (!sprite_area_type)
    {
      static const GtkTypeInfo sprite_area_type_info =
      {
	"GmbSpriteArea",
	sizeof (GmbSpriteArea),
	sizeof (GmbSpriteAreaClass),
	(GtkClassInitFunc) gmb_sprite_area_class_init,
	(GtkObjectInitFunc) gmb_sprite_area_init,
	NULL, NULL,
	(GtkClassInitFunc) NULL
      };
      sprite_area_type = gtk_type_unique (GTK_TYPE_WIDGET,
    					  &sprite_area_type_info);
    }
  return sprite_area_type;
}

GtkWidget*
gmb_sprite_area_new (void)
{
  return GTK_WIDGET (gtk_type_new (gmb_sprite_area_get_type ()));
}

void
gmb_sprite_area_add_sprite_top (GmbSpriteArea * sa, GmbSprite * sprite, gint16 x, gint16 y)
{
  g_return_if_fail (IS_GMB_SPRITE_AREA (sa));
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  gtk_object_ref (GTK_OBJECT (sprite));
  gtk_object_sink (GTK_OBJECT (sprite));

  if (sprite->host)
    gmb_sprite_unparent (sprite);

  sprite->host = sa;

  if (sa->priv->top)
    sa->priv->top->priv->next = sprite;
  else
    sa->priv->bottom = sprite;

  sprite->priv->prev = sa->priv->top;
  sprite->priv->next = NULL;

  sa->priv->top = sprite;

  sprite->area.x = x;
  sprite->area.y = y;

  gmb_sprite_area_setup_added_sprite (sprite);
}

void
gmb_sprite_area_add_sprite_bottom (GmbSpriteArea * sa, GmbSprite * sprite, gint16 x, gint16 y)
{
  g_return_if_fail (IS_GMB_SPRITE_AREA (sa));
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  gtk_object_ref (GTK_OBJECT (sprite));
  gtk_object_sink (GTK_OBJECT (sprite));

  sprite->host = sa;

  if (sa->priv->bottom)
    sa->priv->bottom->priv->prev = sprite;
  else
    sa->priv->top = sprite;

  sprite->priv->prev = NULL;
  sprite->priv->next = sa->priv->bottom;

  sa->priv->bottom = sprite;

  sprite->area.x = x;
  sprite->area.y = y;

  gmb_sprite_area_setup_added_sprite (sprite);
}

void
gmb_sprite_raise (GmbSprite * sprite)
{
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  if (sprite->priv->next)
    {
      GmbSprite * ex_next;
      GmbSprite * ex_prev;

      ex_next = sprite->priv->next;
      ex_prev = sprite->priv->prev;

      sprite->priv->next = ex_next->priv->next;
      sprite->priv->prev = ex_next;

      ex_next->priv->next = sprite;
      ex_next->priv->prev = ex_prev;
    }

  if (sprite->host)
    {
      if (sprite->priv->next == NULL)
        sprite->host->priv->top = sprite;

      sprite->host->priv->need_repick = TRUE;
      gmb_sprite_area_request_update (sprite->host, &sprite->area);
    }
}

void
gmb_sprite_lower (GmbSprite * sprite)
{
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  if (sprite->priv->prev)
    {
      GmbSprite * ex_next;
      GmbSprite * ex_prev;

      ex_next = sprite->priv->next;
      ex_prev = sprite->priv->prev;

      sprite->priv->next = ex_prev;
      sprite->priv->prev = ex_prev->priv->prev;

      ex_prev->priv->prev = sprite;
      ex_prev->priv->next = ex_next;
    }

  if (sprite->host)
    {
      if (sprite->priv->prev == NULL)
        sprite->host->priv->bottom = sprite;

      sprite->host->priv->need_repick = TRUE;
      gmb_sprite_area_request_update (sprite->host, &sprite->area);
    }
}

void
gmb_sprite_raise_to_top (GmbSprite * sprite)
{
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  if (sprite->host && sprite->host->priv->top == sprite)
    return;

  while (sprite->priv->next)
    {
      GmbSprite * ex_next;
      GmbSprite * ex_prev;

      ex_next = sprite->priv->next;
      ex_prev = sprite->priv->prev;

      sprite->priv->next = ex_next->priv->next;
      sprite->priv->prev = ex_next;

      ex_next->priv->next = sprite;
      ex_next->priv->prev = ex_prev;
    }

  if (sprite->host)
    {
      sprite->host->priv->top = sprite;
      sprite->host->priv->need_repick = TRUE;
      gmb_sprite_area_request_update (sprite->host, &sprite->area);
    }
}

void
gmb_sprite_lower_to_bottom (GmbSprite * sprite)
{
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  if (sprite->host && sprite->host->priv->bottom == sprite)
    return;

  while (sprite->priv->prev)
    {
      GmbSprite * ex_next;
      GmbSprite * ex_prev;

      ex_next = sprite->priv->next;
      ex_prev = sprite->priv->prev;

      sprite->priv->next = ex_prev;
      sprite->priv->prev = ex_prev->priv->prev;

      ex_prev->priv->prev = sprite;
      ex_prev->priv->next = ex_next;
    }

  if (sprite->host)
    {
      sprite->host->priv->bottom = sprite;
      sprite->host->priv->need_repick = TRUE;
      gmb_sprite_area_request_update (sprite->host, &sprite->area);
    }
}

void
gmb_sprite_set_position (GmbSprite * sprite, gint16 new_x, gint16 new_y)
{
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  if (!sprite->host)
    {
      sprite->area.x = new_x;
      sprite->area.y = new_y;
    }
  else
    {
      if (ABS (new_x - sprite->area.x) > sprite->area.width &&
	  ABS (new_y - sprite->area.y) > sprite->area.height)
	{
          gmb_sprite_area_request_update (sprite->host, &sprite->area);
          sprite->area.x = new_x;
          sprite->area.y = new_y;
          gmb_sprite_area_request_update (sprite->host, &sprite->area);
	}
      else
	{
          GdkRectangle update;

    	  update.x = MIN (sprite->area.x, new_x);
          update.y = MIN (sprite->area.y, new_y);
          update.width = sprite->area.width + ABS (new_x - sprite->area.x);
          update.height = sprite->area.height + ABS (new_y - sprite->area.y);

          sprite->area.x = new_x;
          sprite->area.y = new_y;

          gmb_sprite_area_request_update (sprite->host, &update);
	}
      sprite->host->priv->need_repick = TRUE;
    }
}

void
gmb_sprite_move (GmbSprite * sprite, gint16 dx, gint16 dy)
{
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  gmb_sprite_set_position (sprite, sprite->area.x + dx, sprite->area.y + dy);
}

void
gmb_sprite_request_resize (GmbSprite * sprite)
{
  guint16 width;
  guint16 height;

  g_return_if_fail (IS_GMB_SPRITE (sprite));

  if (GMB_SPRITE_CLASS (sprite->object.klass)->size_request)
    (* GMB_SPRITE_CLASS (sprite->object.klass)->size_request) (sprite, &width, &height);
  else
    {
      width = 0;
      height = 0;
    }

  if (!sprite->host)
    {
      sprite->area.width = width;
      sprite->area.height = height;
    }
  else
    {
      GdkRectangle update;

      update.x = sprite->area.x;
      update.y = sprite->area.y;
      update.width = MAX (width, sprite->area.width);
      update.height = MAX (height, sprite->area.height);

      sprite->area.width = width;
      sprite->area.height = height;

      gmb_sprite_area_request_update (sprite->host, &update);
      sprite->host->priv->need_repick = TRUE;
    }
}

void
gmb_sprite_area_request_update (GmbSpriteArea * sa, GdkRectangle * area)
{
  GdkRectangle tmp;
  GtkWidget * widget;

  g_return_if_fail (IS_GMB_SPRITE_AREA (sa));
  g_return_if_fail (area != NULL);

  widget = (GtkWidget *) sa;

  if (area->x >= widget->allocation.width ||
      area->y >= widget->allocation.height ||
      area->x + area->width <= 0 ||
      area->y + area->height <= 0 ||
      area->width == 0 || area->height == 0)
    return;

  tmp = *area;

  tmp.x = MAX (area->x, 0);
  tmp.y = MAX (area->y, 0);
  tmp.width = area->x + area->width;
  tmp.width = (tmp.width > widget->allocation.width ?
	       widget->allocation.width - tmp.x : tmp.width - tmp.x);
  tmp.height = area->y + area->height;
  tmp.height = (tmp.height > widget->allocation.height ?
	       widget->allocation.height - tmp.y : tmp.height - tmp.y);

  gmb_sprite_area_add_area_to_queue (sa, &tmp);

  if (!sa->priv->frozen && !sa->priv->idle_id)
    sa->priv->idle_id = gtk_idle_add ((GtkFunction) gmb_sprite_area_idle_func, sa);
}

void
gmb_sprite_area_freeze (GmbSpriteArea * sa)
{
  g_return_if_fail (IS_GMB_SPRITE_AREA (sa));

  if (!sa->priv->frozen)
    {
      sa->priv->frozen = 1;
      if (sa->priv->idle_id)
	{
          gtk_idle_remove (sa->priv->idle_id);
          sa->priv->idle_id = 0;
	}
    }
}

void
gmb_sprite_area_thaw (GmbSpriteArea * sa)
{
  g_return_if_fail (IS_GMB_SPRITE_AREA (sa));

  if (sa->priv->frozen)
    {
      sa->priv->frozen = 0;
      sa->priv->current = NULL;
      sa->priv->need_repick = TRUE;

      if (!sa->priv->idle_id)
	sa->priv->idle_id = gtk_idle_add ((GtkFunction) gmb_sprite_area_idle_func, sa);
    }
}

void
gmb_sprite_area_flush (GmbSpriteArea * sa)
{
  g_return_if_fail (IS_GMB_SPRITE_AREA (sa));

  if (sa->priv->idle_id)
    gtk_idle_remove (sa->priv->idle_id);

  gmb_sprite_area_idle_func (sa);
}

void
gmb_sprite_unparent (GmbSprite * sprite)
{
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  if (sprite->host)
    {
      if (sprite->priv->prev)
        sprite->priv->prev->priv->next = sprite->priv->next;
      if (sprite->priv->next)
        sprite->priv->next->priv->prev = sprite->priv->prev;
      if (sprite->host->priv->top == sprite)
        sprite->host->priv->top = sprite->priv->prev;
      if (sprite->host->priv->bottom == sprite)
        sprite->host->priv->bottom = sprite->priv->next;

      if (sprite == sprite->host->priv->current)
        {
	  sprite->host->priv->current = NULL;
	  sprite->host->priv->need_repick = TRUE;
	}

      gmb_sprite_area_request_update (sprite->host, &sprite->area);
      sprite->host = NULL;
      sprite->priv->next = NULL;
      sprite->priv->prev = NULL;
    }
}

void gmb_sprite_area_foreach (GmbSpriteArea * sa,
			      GmbSpriteFunc f,
			      gpointer data)
{
  GmbSprite * sprite;
  GmbSprite *next_sprite;

  g_return_if_fail (f != NULL);
  g_return_if_fail (IS_GMB_SPRITE_AREA (sa));

  sprite = sa->priv->bottom;
  while (sprite)
    {
      next_sprite = sprite->priv->next;
      (* f) (sprite, data);
      sprite = next_sprite;
    }
}

void
gmb_sprite_show (GmbSprite * sprite)
{
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  if (sprite->object.flags & GMB_SPRITE_VISIBLE)
    return;
  else
    {
      sprite->object.flags |= GMB_SPRITE_VISIBLE;
      if (sprite->host)
        {
	  sprite->host->priv->need_repick = TRUE;
	  gmb_sprite_area_request_update (sprite->host, &sprite->area);
	}
    }
}

void
gmb_sprite_hide (GmbSprite * sprite)
{
  g_return_if_fail (IS_GMB_SPRITE (sprite));

  if (sprite->object.flags & GMB_SPRITE_VISIBLE)
    {
      sprite->object.flags &= ~GMB_SPRITE_VISIBLE;
      if (sprite->host)
        {
	  if (sprite->host->priv->current == sprite)
	    {
	      sprite->host->priv->current = NULL;
	      sprite->host->priv->need_repick = TRUE;
	    }

	  gmb_sprite_area_request_update (sprite->host, &sprite->area);
	}
    }
}
